import collections

from typing import Union

import numpy as np

from numpy.random import RandomState
from scipy.integrate import solve_ivp

from ._seed import get_seed


def _mg_eq(xt, xtau, a=0.2, b=0.1, n=10):
    """
    Mackey-Glass time delay diffential equation, at values x(t) and x(t-tau).

    """
    return -b*xt + a*xtau / (1+xtau**n)


def _mg_rk4(xt, xtau, a, b, n, h=1.0):
    """
    Runge-Kuta method (RK4) for Mackey-Glass timeseries discretization.

    """
    k1 = h * _mg_eq(xt, xtau, a, b, n)
    k2 = h * _mg_eq(xt + 0.5*k1, xtau, a, b, n)
    k3 = h * _mg_eq(xt + 0.5*k2, xtau, a, b, n)
    k4 = h * _mg_eq(xt + k3, xtau, a, b, n)

    return xt + k1/6 + k2/3 + k3/3 + k4/6


def henon_map(n_timesteps: int,
              a: float = 1.4,
              b: float = 0.3,
              x0: Union[list, np.ndarray] = [0.0, 0.0]) -> np.ndarray:
    """Hénon map discrete timeseries [#]_ [#]_:

    .. math::

        x(n+1) &= 1 - ax(n)^2 + y(n)\\\\
        y(n+1) &= bx(n)

    Parameters
    ----------
        n_timesteps : int
            Number of timesteps to generate.
        a : float, optional
            :math:`a` parameter of the system.
            By default, equal to 1.4.
        b : float, optional
            :math:`b` parameter of the system.
            By default, equal to 0.3.
        x0 : list or numpy.ndarray, optional
            Initial conditions of the system.
            By default, equal to [0.0, 0.0].

    Returns
    -------
        numpy.ndarray :
            Hénon map discrete timeseries.

    References
    ----------
        .. [#] M. Hénon, ‘A two-dimensional mapping with a strange
               attractor’, Comm. Math. Phys., vol. 50, no. 1, pp. 69–77, 1976.
        .. [#] `Hénon map <https://en.wikipedia.org/wiki/H%C3%A9non_map>`_
               on Wikipédia

    """
    states = np.zeros((n_timesteps, 2))
    states[0] = np.asarray(x0)

    for i in range(1, n_timesteps):
        states[i][0] = 1 - a*states[i-1][0]**2 + states[i-1][1]
        states[i][1] = b*states[i-1][0]

    return states


def logistic_map(n_timesteps: int,
                 r: float = 3.9,
                 x0: float = 0.5) -> np.ndarray:
    """Logistic map discrete timeseries [#]_ [#]_:

    .. math::

        x(n+1) = rx(n)(1-x(n))

    Parameters
    ----------
        n_timesteps : int
            Number of timesteps to generate.
        r : float, optional
            :math:`r` parameter of the system.
            By default, equal to 3.9.
        x0 : float, optional
            Initial condition of the system.
            By default, equal to 0.5.

    Returns
    -------
        numpy.ndarray :
            Logistic map discrete timeseries.

    References
    ----------
        .. [#] R. M. May, ‘Simple mathematical models with very
               complicated dynamics’, Nature, vol. 261, no. 5560,
               Art. no. 5560, Jun. 1976, doi: 10.1038/261459a0.

        .. [#] `Logistic map <https://en.wikipedia.org/wiki/Logistic_map>`_
               on Wikipédia
    """
    if r > 0 and 0 < x0 < 1:
        X = np.zeros(n_timesteps)
        X[0] = x0

        for i in range(1, n_timesteps):
            X[i] = r * X[i-1] * (1-X[i-1])

        return X
    elif r <= 0:
        raise ValueError("r should be positive.")
    else:
        raise ValueError("Initial condition x0 should be in ]0;1[.")


def lorenz(n_timesteps: int,
           rho: float = 28.0,
           sigma: float = 10.0,
           beta: float = 8.0/3.0,
           x0: Union[list, np.ndarray] = [1.0, 1.0, 1.0],
           h: float = 0.03) -> np.ndarray:
    """Lorenz attractor timeseries [#]_ [#]_:

    .. math::

        \\frac{dx}{dt} &= \\sigma (y-x) \\\\
        \\frac{dy}{dt} &= x(\\rho - z) - y \\\\
        \\frac{dz}{dt} &= xy - \\beta z

    Parameters
    ----------
        n_timesteps : int
            Number of timesteps to generate.
        rho : float, optional
            :math:`\\rho` parameter of the system.
            By default, equal to 28.0.
        sigma : float, optional
            :math:`\\sigma` parameter of the system.
            By default, equal to 10.0.
        beta : float, optional
            :math:`\\beta` parameter of the system.
            By default, equal to :math:`\\frac{8}{3}`.
        x0 : list or numpy.ndarray, optional
            Initial conditions of the system.
            By default, equal to [1.0, 1.0, 1.0].
        h : float, optional
            Controls the continuous time delta between two
            discrete timesteps.
            By default, equal to 0.03.

    Returns
    -------
    np.ndarray
        Lorenz attractor timeseries.

    References
    ----------

        .. [#] E. N. Lorenz, ‘Deterministic Nonperiodic Flow’,
               Journal of the Atmospheric Sciences, vol. 20, no. 2,
               pp. 130–141, Mar. 1963,
               doi: 10.1175/1520-0469(1963)020<0130:DNF>2.0.CO;2.
        .. [#] `Lorenz system <https://en.wikipedia.org/wiki/Lorenz_system>`_
               on Wikipedia.
    """
    def lorenz_diff(t, state):
        x, y, z = state
        return sigma * (y - x), x * (rho - z) - y, x * y - beta * z

    sol = solve_ivp(lorenz_diff,
                    y0=x0,
                    t_span=(0.0, n_timesteps*h),
                    dense_output=True)

    t = np.arange(0.0, n_timesteps * h, h)

    return sol.sol(t).T


def mackey_glass(n_timesteps: int,
                 tau: int = 17,
                 a: float = 0.2,
                 b: float = 0.1,
                 n: int = 10,
                 x0: float = 1.2,
                 h: float = 1.0,
                 seed: Union[int, RandomState] = None) -> np.ndarray:
    """Mackey-Glass timeseries [#]_ [#]_, computed from the Mackey-Glass
    delayed differential equation:

    .. math::

        \\frac{x}{t} = \\frac{ax(t-\\tau)}{1+x(t-\\tau)^n} - bx(t)

    Parameters
    ----------
        n_timesteps : int
            Number of timesteps to compute.
        tau : int, optional
            Time delay :math:`\\tau` of Mackey-Glass equation.
            By defaults, equal to 17. Other values can
            change the choatic behaviour of the timeseries.
        a : float, optional
            :math:`a` parameter of the equation.
            By default, equal to 0.2.
        b : float, optional
            :math:`b` parameter of the equation.
            By default, equal to 0.1.
        n : int, optional
            :math:`n` parameter of the equation.
            By default, equal to 10.
        x0 : float, optional
            Initial condition of the timeseries.
            By default, equal to 1.2.
        h : float, optional
            Time delta for the Runge-Kuta method. Can be assimilated
            to the number of discrete point computed per timestep.
            By default, equal to 1.0.
        seed : int or RandomState
            Random state seed for reproducibility.

    Returns
    -------
        np.ndarray
            Mackey-Glass timeseries.

    Note
    ----
        As Mackey-Glass is defined by delayed time differential equations,
        the first timesteps of the timeseries can't be initialized at 0
        (otherwise, the first steps of computation involving these
        not-computed-yet-timesteps would yield inconsistent results).
        A random number generator is therefore used to produce random
        initial timesteps based on the value of the initial condition
        passed as parameter. A default seed is hard-coded to ensure
        reproducibility in any case. It can be changed with the
        :py:func:`reservoirpy.datasets.seed` function.

    References
    ----------
        .. [#] M. C. Mackey and L. Glass, ‘Oscillation and chaos in physiological
               control systems’, Science, vol. 197, no. 4300, pp. 287–289, Jul. 1977,
               doi: 10.1126/science.267326.
        .. [#] `Mackey-Glass equations
                <https://en.wikipedia.org/wiki/Mackey-Glass_equations>`_
                on Wikipedia.

    """
    # a random state is needed as the method used to discretize
    # the timeseries needs to use randomly generated initial steps
    # based on the initial condition passed as parameter.
    if isinstance(seed, RandomState):
        rs = seed
    elif seed is not None:
        rs = RandomState(seed)
    else:
        rs = RandomState(get_seed())

    # generate random first step based on the value
    # of the initial condition
    history_length = int(np.floor(tau/h))
    history = collections.deque(x0 * np.ones(history_length)
                                + 0.2 * (rs.rand(history_length) - 0.5))
    xt = x0

    X = np.zeros(n_timesteps)

    for i in range(0, n_timesteps):
        X[i] = xt

        if tau == 0:
            xtau = 0.0
        else:
            xtau = history.popleft()
            history.append(xt)

        xth = _mg_rk4(xt, xtau, a=a, b=b, n=n)

        xt = xth

    return X


def multiscroll(n_timesteps: int,
                a: float = 40.0,
                b: float = 3.0,
                c: float = 28.0,
                x0: Union[list, np.ndarray] = [-0.1, 0.5, -0.6],
                h: float = 0.01) -> np.ndarray:
    """Double scroll attractor timeseries [#]_ [#]_,
    a particular case of multiscroll attractor timeseries.

    .. math::

        \\frac{dx}{dt} &= a(y - x) \\\\
        \\frac{dy}{dt} &= (c - a)x - xz + cy \\\\
        \\frac{dz}{dt} &= xy - bz

    Parameters
    ----------
        n_timesteps : int
            Number of timesteps to generate.
        a : float, optional
            :math:`a` parameter of the system.
            By default, equal to 40.
        b : float, optional
            :math:`b` parameter of the system.
            By default, equal to 3.
        c : float, optional
            :math:`c` parameter of the system.
            By default, equal to 28`.
        x0 : list or numpy.ndarray, optional
            Initial conditions of the system.
            By default, equal to [-0.1, 0.5, -0.6].
        h : float, optional
            Controls the continuous time delta between two
            discrete timesteps.
            By default, equal to 0.01.

    Returns
    -------
        numpy.ndarray :
            Multiscroll attractor timeseries.

    References
    ----------
        .. [#] G. Chen and T. Ueta, ‘Yet another chaotic attractor’,
               Int. J. Bifurcation Chaos, vol. 09, no. 07, pp. 1465–1466,
               Jul. 1999, doi: 10.1142/S0218127499001024.
        .. [#] `Chen double scroll attractor
               <https://en.wikipedia.org/wiki/Multiscroll_attractor#Chen_attractor>`_
               on Wikipedia.

    """
    def multiscroll_diff(t, state):
        x, y, z = state
        dx = a * (y - x)
        dy = (c - a)*x - x*z + c*y
        dz = x*y - b*z
        return dx, dy, dz

    t = np.arange(0.0, n_timesteps * h, h)

    sol = solve_ivp(multiscroll_diff,
                    y0=x0,
                    t_span=(0.0, n_timesteps*h),
                    dense_output=True)

    return sol.sol(t).T


def rabinovich_fabrikant(n_timesteps: int,
                         gamma: float = 0.89,
                         alpha: float = 1.1,
                         x0: Union[list, np.ndarray] = [-1, 0, 0.5],
                         h: float = 0.05) -> np.ndarray:
    """Rabinovitch-Fabrikant system [#]_ [#]_ timeseries.

    .. math::

        \\frac{dx}{dt} &= y(z - 1 + x^2) + \\gamma x \\\\
        \\frac{dy}{dt} &= x(3z + 1 - x^2) + \\gamma y \\\\
        \\frac{dz}{dt} &= -2z(\\alpha + xy)

    Parameters
    ----------
        n_timesteps : int
            Number of timesteps to generate.
        alpha : float, optional
            :math:`\\alpha` parameter of the system.
            By default, equal to 1.1.
        gamma : float, optional
            :math:`\\gamma` parameter of the system.
            By default, equal to 0.89.
        x0 : list or numpy.ndarray, optional
            Initial conditions of the system.
            By default, equal to [-1, 0, 0.5].
        h : float, optional
            Controls the continuous time delta between two
            discrete timesteps.
            By default, equal to 0.05.

    Returns
    -------
        numpy.ndarray
            Rabinovitch-Fabrikant system timeseries.

    References
    ----------
        .. [#] M. I. Rabinovich and A. L. Fabrikant,
               ‘Stochastic self-modulation of waves in
               nonequilibrium media’, p. 8, 1979.
        .. [#] `Rabinovich-Fabrikant equations
               <https://en.wikipedia.org/wiki/Rabinovich%E2%80%93Fabrikant_equations>`_
               on Wikipedia.

    """
    def rabinovich_fabrikant_diff(t, state):
        x, y, z = state
        dx = y*(z - 1 + x**2) + gamma*x
        dy = x*(3*z + 1 - x**2) + gamma*y
        dz = -2*z*(alpha + x*y)
        return dx, dy, dz

    t = np.arange(0.0, n_timesteps*h, h)

    sol = solve_ivp(rabinovich_fabrikant_diff,
                    y0=x0,
                    t_span=(0.0, n_timesteps*h),
                    dense_output=True)

    return sol.sol(t).T



from ._version import __version__

from ._utils import load

from . import mat_gen
from . import observables
from . import regression_models
from . import activationsfunc
from . import hyper

from ._esn import ESN
from ._esn_online import ESNOnline

__all__ = [
    "__version__",
    "ESN", "ESNOnline",
    "mat_gen", "activationsfunc",
    "observables", "regression_models",
    "hyper", "load"
]



import pytest
import numpy as np

from numpy.testing import assert_allclose

from reservoirpy import datasets


@pytest.mark.parametrize("dataset_func", [
    datasets.henon_map,
    datasets.logistic_map,
    datasets.lorenz,
    datasets.mackey_glass,
    datasets.multiscroll,
    datasets.rabinovich_fabrikant
])
def test_generation(dataset_func):

    timesteps = 2000
    X = dataset_func(timesteps)

    assert isinstance(X, np.ndarray)
    assert len(X) == timesteps


@pytest.mark.parametrize("dataset_func", [
    datasets.mackey_glass
])
def test_seed(dataset_func):
    x1 = dataset_func(200)
    x2 = dataset_func(200)

    assert_allclose(x1, x2)


@pytest.mark.parametrize("dataset_func", [
    datasets.mackey_glass
])
def test_reseed(dataset_func):

    s = datasets.get_seed()
    assert s == datasets._seed._DEFAULT_SEED

    x1 = dataset_func(200)

    datasets.set_seed(1234)
    assert datasets._seed._DEFAULT_SEED == 1234
    assert datasets.get_seed() == 1234

    x2 = dataset_func(200)

    assert (np.abs(x1 - x2) > 1e-3).sum() > 0



"""Utility module for optimization of hyperparameters
using *hyperopt*.

Note
----

This module is meant to be used alongside *hyperopt*
and *Matplotlib* packages, which are not installed
with ReservoirPy by default. Before using the
:py:mod:`reservoirpy.hyper` module, consider installing
these packages if they are not already installed.
"""
from ._hypersearch import research
from ._hyperplot import plot_hyperopt_report

__all__ = ["research", "plot_hyperopt_report"]



import numpy as np

from scipy import linalg
from scipy.sparse import issparse
from scipy.sparse.linalg import eigs


def spectral_radius(W, maxiter: int = None) -> float:
    """Compute the spectral radius of a matrix `W`.

    Spectral radius is defined as the maximum absolute
    eigenvalue of `W`.

    Parameters
    ----------
    W : numpy.ndarray or scipy.sparse matrix
        Matrix from which the spectral radius will
        be computed.

    maxiter : int, optional
        Maximum number of Arnoldi update iterations allowed.
        By default, equal to `W.shape[0] * 20`.
        See `Scipy documentation <https://docs.scipy.org/
        doc/scipy/reference/generated/scipy.sparse.linalg.eigs.html>`_
        for more informations.

    Returns
    -------
    float
        Spectral radius of `W`.

    Raises
    ------
    ArpackNoConvergence
        When computing spectral radius on large
        sparse matrices, it is possible that the
        Fortran ARPACK algorithmn used to compute
        eigenvalues don't converge towards precise
        values. To avoid this problem, set the `maxiter`
        parameter to an higher value. Be warned that
        this may drastically increase the computation
        time.

    """
    if issparse(W):
        if maxiter is None:
            maxiter = W.shape[0] * 20

        return max(abs(eigs(W,
                            k=1,
                            which='LM',
                            maxiter=maxiter,
                            return_eigenvectors=False)))

    return max(abs(linalg.eig(W)[0]))


def compute_error_NRMSE(teacher_signal, predicted_signal, verbose=False):
    """ Computes Normalized Root-Mean-Squarred Error between a teacher signal and a predicted signal
    Return the errors in this order: nmrse mean, nmrse max-min, rmse, mse.
    By default, only NMRSE mean should be considered as a general measure to be
    compared for different datasets.

    For more information, see:
    - Mean Squared Error https://en.wikipedia.org/wiki/Mean_squared_error
    - Root Mean Squared Error https://en.wikipedia.org/wiki/Root-mean-square_deviation for more info
    """
    errorLen = len(predicted_signal[:])
    mse = np.mean((teacher_signal - predicted_signal)**2)
    rmse = np.sqrt(mse)
    nmrse_mean = abs(rmse / np.mean(predicted_signal[:])) # Normalised RMSE (based on mean)
    nmrse_maxmin = rmse / abs(np.max(predicted_signal[:]) - np.min(predicted_signal[:])) # Normalised RMSE (based on max - min)
    if verbose:
        print("Errors computed over %d time steps" % (errorLen))
        print("\nMean Squared error (MSE):\t\t%.4e" % (mse) )
        print("Root Mean Squared error (RMSE):\t\t%.4e\n" % rmse )
        print("Normalized RMSE (based on mean):\t%.4e" % (nmrse_mean) )
        print("Normalized RMSE (based on max - min):\t%.4e" % (nmrse_maxmin) )
    return nmrse_mean, nmrse_maxmin, rmse, mse



"""*Matplotlib* wrapper tools for optimization of
hyperparameters results display and analysis.

"""
import os
import math
import json

from os import path

import numpy as np


def _get_results(exp):
    report_path = path.join(exp, "results")
    results = []
    for file in os.listdir(report_path):
        if path.isfile(path.join(report_path, file)):
            with open(path.join(report_path, file), 'r') as f:
                results.append(json.load(f))
    return results


def _outliers_idx(values, max_deviation):
    mean = values.mean()

    dist = abs(values - mean)
    corrected = dist < mean + max_deviation
    return corrected


def _logscale_plot(ax, xrange, yrange, base=10):

    from matplotlib import ticker

    if xrange is not None:
        ax.xaxis.set_minor_formatter(ticker.LogFormatter())
        ax.xaxis.set_major_formatter(ticker.LogFormatter())
        ax.set_xscale("log", base=base)
        ax.set_xlim([np.min(xrange), np.max(xrange)])
    if yrange is not None:
        ax.yaxis.set_minor_formatter(ticker.LogFormatter())
        ax.yaxis.set_major_formatter(ticker.LogFormatter())
        ax.set_yscale("log", base=base)
        ax.set_ylim([yrange.min() - 0.1*yrange.min(),
                     yrange.max() + 0.1*yrange.min()])


def _scale(x):
    return (x - x.min()) / (x.ptp())


def _cross_parameter_plot(ax, values, scores, loss, smaxs, cmaxs,
                          lmaxs, p1, p2, log1, log2, cat1, cat2):

    X = values[p2].copy()
    Y = values[p1].copy()

    to_log = []
    if log1 and not cat2:
        to_log.append(X)
    else:
        to_log.append(None)
        if cat2:
            ax.margins(x=0.05)
    if log1 and not cat1:
        to_log.append(Y)
    else:
        to_log.append(None)

    _logscale_plot(ax, *to_log)
    # if log2 and not cat2:
    #     logscale_plot(ax, None, Y)

    ax.tick_params(axis='both', which='both')
    ax.tick_params(axis='both', labelsize="xx-small")
    ax.grid(True, which="both", ls="-", alpha=0.75)

    ax.set_xlabel(p2)
    ax.set_ylabel(p1)

    sc_l = ax.scatter(X[lmaxs], Y[lmaxs], scores[lmaxs]*100,
                      c=loss[lmaxs], cmap="inferno")
    sc_s = ax.scatter(X[smaxs], Y[smaxs], scores[smaxs]*100,
                      c=cmaxs, cmap="YlGn")
    sc_m = ax.scatter(X[~(lmaxs)], Y[~(lmaxs)], scores[~(lmaxs)]*100,
                      color="red")

    return sc_l, sc_s, sc_m


def _loss_plot(ax, values, scores, loss, smaxs, cmaxs, lmaxs,
               p, log, categorical, legend, loss_behaviour):

    X = values[p].copy()

    if log and not(categorical):
        _logscale_plot(ax, X, loss)
    else:
        _logscale_plot(ax, None, loss)
        if categorical:
            ax.margins(x=0.05)

    ax.set_xlabel(p)
    ax.set_ylabel("loss")

    ax.tick_params(axis='both', which='both')
    ax.tick_params(axis='both', labelsize="xx-small")
    ax.grid(True, which="both", ls="-", alpha=0.75)

    sc_l = ax.scatter(X[lmaxs], loss[lmaxs], scores[lmaxs]*100, color="orange")
    sc_s = ax.scatter(X[smaxs], loss[smaxs], scores[smaxs]*100, c=cmaxs, cmap="YlGn")
    if loss_behaviour == "min":
        sc_m = ax.scatter(X[~(lmaxs)], [loss.min()]*np.sum(~lmaxs),
                          scores[~(lmaxs)]*100, color="red", label="Loss min.")
    else:
        sc_m = ax.scatter(X[~(lmaxs)], [loss.max()]*np.sum(~lmaxs),
                          scores[~(lmaxs)]*100, color="red", label="Score max.")

    if legend:
        ax.legend()

    return sc_l, sc_s, sc_m


def _parameter_violin(ax, values, scores, loss,
                      smaxs, cmaxs, p, log, legend):

    import matplotlib.pyplot as plt

    y = values[p].copy()[smaxs]
    all_y = values[p].copy()

    if log:
        y = np.log10(y)
        all_y = np.log10(all_y)

    ax.get_yaxis().set_ticks([])
    ax.tick_params(axis='x', which='both')

    ax.xaxis.set_major_locator(plt.MultipleLocator(1))

    def format_func(value, tick_number):
        return "$10^{"+str(int(np.floor(value)))+"}$"

    ax.xaxis.set_major_formatter(plt.FuncFormatter(format_func))

    ax.set_xlabel(p)
    ax.grid(True, which="both", ls="-", alpha=0.75)
    ax.set_xlim([np.min(all_y), np.max(all_y)])

    violin = ax.violinplot(y, vert=False, showmeans=False, showextrema=False)

    for pc in violin['bodies']:
        pc.set_facecolor('forestgreen')
        pc.set_edgecolor('white')

    quartile1, medians, quartile3 = np.percentile(y, [25, 50, 75])

    ax.scatter(medians, 1, marker='o', color='orange', s=30,
               zorder=4, label="Median")
    ax.hlines(1, quartile1, quartile3, color='gray',
              linestyle='-', lw=4, label="Q25/Q75")
    ax.vlines(y.mean(), 0.5, 1.5, color='blue', label="Mean")

    ax.scatter(np.log10(values[p][scores.argmax()]), 1, color="red",
               zorder=5, label="Best score")
    ax.scatter(y, np.ones_like(y), c=cmaxs, cmap="YlGn",
               alpha=0.5, zorder=3)

    if legend:
        ax.legend(loc=2)


def _parameter_bar(ax, values, scores, loss, smaxs, cmaxs, p, categories):

    y = values[p].copy()[smaxs]

    ax.set_xlabel(p)
    ax.grid(True, which="both", ls="-", alpha=0.75)

    heights = []
    for p in categories:
        heights.append(y.tolist().count(p))

    ax.bar(x=categories, height=heights, color="forestgreen", alpha=0.3)


def plot_hyperopt_report(exp, params, metric='loss', loss_metric="loss",
                         loss_behaviour="min", not_log=None, categorical=None,
                         max_deviation=None, title=None):
    """Cross paramater scatter plot of hyperopt trials.

    Note
    ----
        Installation of Matplotlib and Seaborn packages
        is required to use this tool.

    Parameters
    ----------
        exp : str or Path
            Report directory storing hyperopt trials results.

        params : list
            Parameters to plot.

        metric : str, optional
            Metric to use as performance measure,
            stored in the hyperopt trials results dictionnaries.
            May be different from loss metric. By default,
            'loss' is used as performance metric.

        loss_metric : str, optional
            Metric to use as an error measure,
            stored in the hyperopt trials results dictionnaries.
            May be different from the default `loss` parameter.

        loss_behaviour : {'min', 'max'}, optional
            How to interpret metric used as main loss in the plot.
            If loss need to be minimized, choose 'min'. If loss
            need to be maximized, choose 'max'. In most cases,
            the loss is an error function that needs to be
            minimized. By default, 'min'.

        not_log : list, optional
            List of parameters to plot with a linear scale. By default,
            all scales are logarithmic.

        categorical : list, optional
            List of parameters to interpret as categorical or
            discrete valued.

        max_deviation : float, optional
            Maximum standard deviation expected from the loss mean.
            Useful to remove extreme outliers that may create odd plots.
            By defautl, all values are kept and plotted.

        title : str, optional
            Optional title for the figure.

    Returns:
        matplotlib.pyplot.figure
            Matplotlib figure object.

    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set(context="paper", style="darkgrid", font_scale=1.5)
    N = len(params)
    not_log = not_log or []

    results = _get_results(exp)

    loss = np.array([r['returned_dict'][loss_metric] for r in results])
    scores = np.array([r['returned_dict'][metric] for r in results])

    if max_deviation is not None:
        not_outliers = _outliers_idx(loss, max_deviation)
        loss = loss[not_outliers]
        scores = scores[not_outliers]
        values = {p: np.array([r['current_params'][p] for r in results])[not_outliers]
                  for p in params}
    else:
        values = {p: np.array([r['current_params'][p] for r in results])
                  for p in params}

    categorical = categorical or []

    for p in categorical:
        values[p] = values[p].astype(str)

    # Sorting for categorical plotting
    all_categorical = [val for p, val in values.items() if p in categorical]
    all_numerical = [val for p, val in values.items() if p not in categorical]

    sorted_idx = np.lexsort((loss, scores, *all_numerical, *all_categorical))

    loss = np.array([loss[i] for i in sorted_idx])
    scores = np.array([scores[i] for i in sorted_idx])

    for p, val in values.items():
        values[p] = np.array([val[i] for i in sorted_idx])

    scores = _scale(scores)

    # loss and f1 values

    if loss_behaviour == "min":
        lmaxs = loss > loss.min()
    else:
        lmaxs = loss < loss.max()

    percent = math.ceil(len(scores) * 0.05)
    smaxs = scores.argsort()[-percent:][::-1]
    cmaxs = _scale(scores[smaxs])

    # gridspecs

    fig = plt.figure(figsize=(15, 19), constrained_layout=True)
    gs = fig.add_gridspec(2, 1, height_ratios=[2/30, 28/30])
    fig.suptitle(f"Hyperopt trials summary - {title}", size=15)

    gs0 = gs[0].subgridspec(1, 3)
    gs1 = gs[1].subgridspec(N + 1, N)

    lbar_ax = fig.add_subplot(gs0[0, 0])
    fbar_ax = fig.add_subplot(gs0[0, 1])
    rad_ax = fig.add_subplot(gs0[0, 2])
    rad_ax.axis('off')

    # plot
    axes = []
    for i, p1 in enumerate(params):
        for j, p2 in enumerate(params):
            ax = fig.add_subplot(gs1[i, j])
            axes.append(ax)
            if p1 == p2:
                sc_l, sc_s, sc_m = _loss_plot(ax, values, scores, loss, smaxs,
                                              cmaxs, lmaxs, p2,
                                              not(p2 in not_log),
                                              p2 in categorical,
                                              (i == 0 and j == 0),
                                              loss_behaviour)
            else:
                sc_l, sc_s, sc_m = _cross_parameter_plot(ax, values,
                                                         scores, loss,
                                                         smaxs, cmaxs,
                                                         lmaxs, p1, p2,
                                                         not(p1 in not_log),
                                                         not(p2 in not_log),
                                                         p1 in categorical,
                                                         p2 in categorical)

    # legends

    handles, labels = sc_l.legend_elements(prop="sizes")
    legend = rad_ax.legend(handles, labels, loc="center left",
                           title=f"Normalized {metric} (%)", mode='expand',
                           ncol=len(labels) // 2 + 1)

    l_cbar = fig.colorbar(sc_l, cax=lbar_ax, ax=axes, orientation="horizontal")
    _ = l_cbar.ax.set_title("Loss value")

    f_cbar = fig.colorbar(sc_s, cax=fbar_ax, ax=axes,
                          orientation="horizontal", ticks=[0, 0.5, 1])
    _ = f_cbar.ax.set_title(f"{metric} best population")
    _ = f_cbar.ax.set_xticklabels(["5% best", "2.5% best", "Best"])

    # violinplots

    legend = True
    for i, p in enumerate(params):
        ax = fig.add_subplot(gs1[-1, i])
        if p in categorical:
            _parameter_bar(ax, values, scores,
                           loss, smaxs, cmaxs, p, sorted(list(set(values[p]))))
        else:
            _parameter_violin(ax, values, scores,
                              loss, smaxs, cmaxs, p, not(p in not_log), legend)
            legend = False
        if legend:
            ax.set_ylabel(f"5% best {metric}\nparameter distribution")
    return fig



"""ReservoirPy toy and life-sized datasets.

Chaotic timeseries
------------------
    Timeseries expressing a chaotic behaviour, generated at will.

    All timeseries defined by differential equations on a
    continuous space are approximated using 4-5th order
    Runge-Kuta method [#]_, either homemade (for Mackey-Glass timeseries)
    or from Scipy `solve_ivp`_ tool.

    Available chaotic attractors:

    **Discrete timeseries**

    - :py:func:`logistic_map`: Logistic map timeseries.

    - :py:func:`henon_map`: Hénon map timeseries.

    **Approximations of continuous timeseries**

    - :py:func:`mackey_glass`: Mackey-Glass delayed differential equations timeseries.

    :- py:func:`lorenz`: Lorenz system timeseries.

    - :py:func:`multiscroll`: Double scroll attractor timeseries.

    - :py:func:`rabinovitch_fabrikant`: Rabinovitch-Fabrikant differential
      equations timeseries.

References
----------
    .. [#] `Runge–Kutta methods
           <https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods>`_
           on Wikipedia.


.. _solve_ivp: https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html
"""
from ._chaos import henon_map
from ._chaos import logistic_map
from ._chaos import lorenz
from ._chaos import multiscroll
from ._chaos import rabinovich_fabrikant
from ._chaos import mackey_glass

from ._seed import set_seed, get_seed


__all__ = [
    "henon_map",
    "logistic_map",
    "lorenz",
    "mackey_glass",
    "multiscroll",
    "rabinovich_fabrikant",
    "set_seed", "get_seed"
]



"""Activations functions for reservoir, feedback and output.
"""
from typing import Callable
from functools import wraps

import numpy as np


def _elementwise(func):
    """Vectorize a function to apply it
    on arrays.
    """
    vect = np.vectorize(func)

    @wraps(func)
    def vect_wrapper(*args, **kwargs):
        u = np.asanyarray(args)
        v = vect(u)
        return v[0]

    return vect_wrapper


def get_function(name: str) -> Callable:  # pragma: no cover
    """Return an activation function from name.

    Parameters
    ----------
    name : str
        Name of the activation function.
        Can be one of {'softmax', 'softplus',
        'sigmoid', 'tanh', 'identity', 'relu'} or
        their respective short names {'smax', 'sp',
        'sig', 'id', 're'}.

    Returns
    -------
    Callable
        An activation function.
    """
    index = {
        "softmax": softmax,
        "softplus": softplus,
        "sigmoid": sigmoid,
        "tanh": tanh,
        "identity": identity,
        "relu": relu,
        "smax": softmax,
        "sp": softplus,
        "sig": sigmoid,
        "id": identity,
        "re": relu,
    }

    if index.get(name) is None:
        raise ValueError(f"Function name must be one of {[k for k in index.keys()]}")
    else:
        return index[name]


def softmax(x: np.ndarray) -> np.ndarray:
    """Softmax activation function:

    .. math::

        y_k = \\frac{e^{x_k}}{\\sum_{i=1}^{n} e^{x_i}}

    Parameters
    ----------
    x : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """
    return np.exp(x) / np.exp(x).sum()


@_elementwise
def softplus(x: np.ndarray) -> np.ndarray:
    """Softplus acctivation function:

    .. math::

        f(x) = \\mathrm{ln}(1 + e^{x})

    Can be used as a smooth version of ReLU.

    Parameters
    ----------
    x : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """
    return np.log(1 + np.exp(x))


@_elementwise
def sigmoid(x: np.ndarray) -> np.ndarray:
    """Sigmoïde activation function:

    .. math::

        f(x) = \\frac{1}{1 + e^{-x}}


    Parameters
    ----------
    x : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """
    if x < 0:
        u = np.exp(x)
        return u / (u + 1)
    return 1 / (1 + np.exp(-x))


def tanh(x: np.ndarray) -> np.ndarray:
    """Hyperbolic tangent activation function:

    .. math::

        f(x) = \\frac{e^x - e^{-x}}{e^x + e^{-x}}

    Parameters
    ----------
    x : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """
    return np.tanh(x)


@_elementwise
def identity(x: np.ndarray) -> np.ndarray:
    """Identity function :

    .. math::

        f(x) = x

    Provided for convenience.

    Parameters
    ----------
    x : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """
    return x


@_elementwise
def relu(x: np.ndarray) -> np.ndarray:
    """ReLU activation function:

    .. math::
        f(x) = x ~~ \\mathrm{if} ~~ x > 0 ~~ \\mathrm{else} ~~ 0

    Parameters
    ----------
    x : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    """
    if x < 0:
        return 0
    return x



"""Simple regression models for readout matrix learning.

This module provides simple linear models that can be used
to compute the readout matrix coefficients with simple
linear regression algorithms, like ridge regularized regression
or any linear model from scikit-learn API.

These models are already packed in the :py:class:`reservoirpy.ESN`
class, and can instanciated by passing them as arguments to the `ESN`
object.

This module defines all models in the form of parametrizable functions.
This functions then return another function that can be used to fit the
model to the data, as shown in example below.

Example
-------

.. code-block:: python

    from reservoirpy.regression_models import ridge_linear_model
    model = ridge_linear_model(ridge=1.e-6)
    ...
    # get some internal states from a reservoir
    # and ground truth vectors (teachers)
    ...
    # compute the readout
    Wout = model(states, teachers)

Using scikit-learn API (for instance, the `Lasso` regression model):

.. code-block:: python

    from sklearn.linear_model import Lasso
    from reservoirpy.regression_models import sklearn_linear_model
    # parametrize the model
    regressor = Lasso(alpha=0.1)
    # create a model function...
    model = sklearn_linear_model(regressor)
    ...
    # or create an ESN and inject the model
    ...
    esn = ESN(..., reg_model=regressor)

In most cases, you won't need to call this module directly. Simply
pass the models to the `ESN` object as parameters.
See the :py:class:`reservoirpy.ESN` documentation for more information.
"""
from typing import Callable

import numpy as np

from scipy import linalg


def sklearn_linear_model(model: Callable):
    """Create a solver from a scikit-learn linear model.

    Parameters
    ----------
    model : sklearn.linear_model instance
        A scikit-learn linear model.
    """
    def linear_model_solving(X, Y):
        # Learning of the model (first row of X, which contains only ones, is removed)
        model.fit(X[1:, :].T, Y.T)

        # linear_model provides Matrix A and Vector b
        # such as A * X[1:, :] + b ~= Ytarget
        A = np.asmatrix(model.coef_)
        b = np.asmatrix(model.intercept_).T

        # Then the matrix W = "[b | A]" statisfies "W * X ~= Ytarget"
        return np.asarray(np.hstack([b, A]))

    return linear_model_solving


def ridge_linear_model(ridge=0., typefloat=np.float32):
    """Create a solver able to perform a linear regression with
    a L2 regularization, also known as Tikhonov or Ridge regression.

    The ridge regression is performed following this equation:

    .. math::

        W_{out} = YX^{T} \cdot (XX^{T} + \mathrm{ridge} \\times \mathrm{Id}_{N})

    where :math:`W_out` is the readout matrix learnt through this regression,
    :math:`X` are the internal states, :math:`Y` are the ground truth vectors,
    and :math:`N` is the internal state dimension (number of units in the
    reservoir).


    Parameters
    ----------
    ridge : float, optional
        Regularization parameter. By default, equal to 0.

    typefloat : numpy.dtype, optional
    """
    def ridge_model_solving(X, Y):
        ridgeid = (ridge*np.eye(X.shape[0])).astype(typefloat)

        return np.dot(np.dot(Y, X.T), linalg.inv(np.dot(X, X.T) + ridgeid))

    return ridge_model_solving


def pseudo_inverse_linear_model():
    """Create a solver able to perform a linear regression
    using an analytical method without regularization.

    The regression is performed following this equation:

    .. math::

        W_{out} = YX^{T}

    where :math:`W_out` is the readout matrix learnt through this regression,
    :math:`X` are the internal states and :math:`Y` are the ground truth vectors.
    """
    def pseudo_inverse_model_solving(X, Y):
        return np.dot(Y, linalg.pinv(X))

    return pseudo_inverse_model_solving



_DEFAULT_SEED = 5555


def get_seed():
    """Return the current random state seed used for dataset
    generation.

    Returns
    -------
    int
        Current seed value.
    """
    global _DEFAULT_SEED
    return _DEFAULT_SEED


def set_seed(s: int):
    """Change the default random seed value.

    This will change the behaviour of the Mackey-Glass
    timeseries generator (see :py:func:`mackey_glass`).

    Parameters
    ----------
    s : int
        A random state generator numerical seed.
    """
    global _DEFAULT_SEED
    _DEFAULT_SEED = s



import os

from setuptools import setup, find_packages

NAME = "reservoirpy"

__version__ = None
version_file = os.path.join("reservoirpy", "_version.py")
with open(version_file) as f:
    exec(f.read())

AUTHOR = "Xavier Hinaut"
AUTHOR_EMAIL = "xavier.hinaut@inria.fr"

MAINTAINERS = "Xavier Hinaut, Nathan Trouvain"
MAINTAINERS_EMAIL = "xavier.hinaut@inria.fr, nathan.trouvain@inria.fr"

DESCRIPTION = ("A simple and flexible code for Reservoir "
               "Computing architectures like Echo State Networks.")

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

URL = "https://github.com/reservoirpy/reservoirpy"
DOWNLOAD_URL = (f"{URL}/v{__version__}.tar.gz")

INSTALL_REQUIRES = [
    "tqdm>=4.43.0",
    "joblib>=0.12",
    "dill>=0.3.0"
    "numpy>=1.15.0",
    "scipy>=1.0.0",
    "joblib>=0.12"
    ]

PROJECT_URLS = {
    'Bug Tracker': 'https://github.com/reservoirpy/reservoirpy/issues',
    'Documentation': 'https://github.com/reservoirpy/reservoirpy',
    'Source Code': URL
}

setup(
    name=NAME,
    version=__version__,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=MAINTAINERS,
    maintainer_email=MAINTAINERS_EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=URL,
    project_urls=PROJECT_URLS,
    download_url=DOWNLOAD_URL,
    packages=find_packages(),
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        ('Topic :: Scientific/Engineering :: '
         'Artificial Intelligence'),
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        ('Programming Language :: Python :: '
         'Implementation :: PyPy')
    ],
    python_requires='>=3.6',
    install_requires=[
        "tqdm>=4.43.0",
        "joblib>=0.12",
        "dill>=0.3.0"
        'numpy>=1.15.0',
        'scipy>=1.0.0',
        'joblib>=0.12',
    ]
)



"""Echo State Networks with online learning rule (FORCE [#]_ learning).

References
----------

    .. [#] D. Sussillo and L. F. Abbott, ‘Generating Coherent
           Patterns of Activity from Chaotic Neural Networks’, Neuron,
           vol. 63, no. 4, pp. 544–557, Aug. 2009,
           doi: 10.1016/j.neuron.2009.07.018.

"""
from typing import Sequence, Callable, Tuple, Union

import numpy as np

from ._utils import _save


class ESNOnline:

    def __init__(self,
                 lr: float,
                 W: np.ndarray,
                 Win: np.ndarray,
                 Wout: np.ndarray,
                 alpha_coef: float = 1e-6,
                 use_raw_input: bool = False,
                 input_bias: bool = True,
                 Wfb: np.ndarray = None,
                 fbfunc: Callable = None,
                 typefloat: np.dtype = np.float64):
        """Echo State Networks using FORCE algorithm as
        an online learning rule.

        The :py:class:`reservoirpy.ESNOnline` implements FORCE, an
        online learning method.
        Online Echo State Networks allow one to:
            - quickly build online ESNs, using the :py:mod:`reservoirpy.mat_gen` module
            to initialize weights,
            - train and test ESNs on the task of your choice in an online fashion, i.e.
            continuously in time.

    Parameters
    ----------
        lr: float
            Leaking rate
        W: np.ndarray
            Reservoir weights matrix
        Win: np.ndarray
            Input weights matrix
        Wout: np.ndarray
            Readout weights matrix
        alpha_coef : float, optional
            Coefficient to scale the inversed state correlation matrix
            used for FORCE learning. By defautl, equal to :math:`1e^{-6}`.
        use_raw_input : bool, optional
            If True, input is used directly when computing output.
            By default, is False.
        input_bias: bool, optional
            If True, will add a constant bias
            to the input vector. By default, True.
        Wfb: np.array, optional
            Feedback weights matrix.
        fbfunc: Callable, optional
            Feedback activation function.
        typefloat: numpy.dtype, optional

    Attributes
    ----------
        Wout: np.ndarray
            Readout matrix
        dim_out: int
            Output dimension
        dim_inp: int
            Input dimension
        N: int
            Number of neuronal units
        state : numpy.ndarray
            Last internal state computed, used to
            store internal dynamics of the reservoir
            over time, to enable online learning.
        output_values : numpy.ndarray
            Last values predicted by the network,
            used to store the last response of the
            reservoir to enable online learning.
        state_corr_inv : numpy.ndarray
            Inverse correlation matrix used in FORCE learning
            algorithm.

    See also
    --------
        reservoirpy.ESN for ESN with offline learning.

    """
        self.W = W
        self.Win = Win
        self.Wfb = Wfb
        self.Wout = Wout

        self.use_raw_inp = use_raw_input

        # check if dimensions of matrices are coherent
        self._autocheck_dimensions()
        self._autocheck_nan()

        # number of neurons
        self.N = self.W.shape[1]

        self.in_bias = input_bias
        # dimension of inputs (not including the bias at 1)
        self.dim_inp = self.Win.shape[1] - 1 if self.in_bias else self.Win.shape[1]

        self.dim_out = Wout.shape[0]
        self.state_size = Wout.shape[1]
        self.output_values = np.zeros((self.dim_out, 1)).astype(typefloat)

        self.typefloat = typefloat
        self.lr = lr # leaking rate

        self.fbfunc = fbfunc
        if self.Wfb is not None and self.fbfunc is None:
            raise ValueError(f"If a feedback matrix is provided, \
                fbfunc must be a callable object, not {self.fbfunc}.")

        # coef used to init state_corr_inv matrix
        self.alpha_coef = alpha_coef

        # Init internal state vector and state_corr_inv matrix
        # (useful if we want to freeze the online learning)
        self.reset_reservoir()
        self.reset_correlation_matrix()

    def __repr__(self):
        trained = True
        if self.Wout is None:
            trained = False
        fb = True
        if self.Wfb is None:
            fb=False
        out = f"ESN(trained={trained}, feedback={fb}, N={self.N}, "
        out += f"lr={self.lr}, input_bias={self.in_bias}, input_dim={self.N})"
        return out

    def _autocheck_nan(self):
        """ Auto-check to see if some important variables do not have a problem (e.g. NAN values). """
        # assert np.isnan(self.W).any() == False, "W matrix should not contain NaN values."
        assert np.isnan(self.Win).any() == False, "Win matrix should not contain NaN values."
        if self.Wfb is not None:
            assert np.isnan(self.Wfb).any() == False, "Wfb matrix should not contain NaN values."

    def _autocheck_dimensions(self):
        """ Auto-check to see if ESN matrices have correct dimensions."""
        # W dimensions check list
        assert len(self.W.shape) == 2, f"W shape should be (N, N) but is {self.W.shape}."
        assert self.W.shape[0] == self.W.shape[1], f"W shape should be (N, N) but is {self.W.shape}."

        # Win dimensions check list
        assert len(self.Win.shape) == 2, f"Win shape should be (N, input) but is {self.Win.shape}."
        err = f"Win shape should be ({self.W.shape[1]}, input) but is {self.Win.shape}."
        assert self.Win.shape[0] == self.W.shape[0], err

        # Wout dimensions check list
        assert len(self.Wout.shape) == 2, f"Wout shape should be (output, nb_states) but is {self.Wout.shape}."
        nb_states = self.Win.shape[1] + self.W.shape[0] + 1 if self.use_raw_inp else self.W.shape[0] + 1
        err = f"Wout shape should be (output, {nb_states}) but is {self.Wout.shape}."
        assert self.Wout.shape[1] == nb_states, err

        # Wfb dimensions check list
        if self.Wfb is not None:
            assert len(self.Wfb.shape) == 2, f"Wfb shape should be (input, output) but is {self.Wfb.shape}."
            err = f"Wfb shape should be ({self.Win.shape[0]}, {self.Wout.shape[0]}) but is {self.Wfb.shape}."
            assert (self.Win.shape[0],self.Wout.shape[0]) == self.Wfb.shape, err

    def _autocheck_io(self,
                      inputs,
                      outputs=None):

        # Check if inputs and outputs are lists
        assert type(inputs) is list, "Inputs should be a list of numpy arrays"
        if outputs is not None:
            assert type(outputs) is list, "Outputs should be a list of numpy arrays"

        # check if Win matrix has coherent dimensions with input dimensions
        inputs_0 = inputs[0]
        if self.in_bias:
            err = f"With bias, Win matrix should be of shape ({self.N}, "
            err += f"{inputs_0.shape[0] + 1}) but is {self.Win.shape}."
            assert self.Win.shape[1] == inputs_0.shape[0] + 1, err
        else:
            err = f"Win matrix should be of shape ({self.N}, "
            err += f"{inputs_0.shape[0]}) but is {self.Win.shape}."
            assert self.Win.shape[1] == inputs_0.shape[0], err

        if outputs is not None:
            # check feedback matrix
            if self.Wfb is not None:
                outputs_0 = outputs[0]
                err = f"With feedback, Wfb matrix should be of shape ({self.N}, "
                err += f"{outputs_0.shape[0]}) but is {self.Wfb.shape}."
                assert outputs_0.shape[0] == self.Wfb.shape[1], err
            # check output weights matrix
            if self.Wout is not None:
                outputs_0 = outputs[0]
                err = f"Wout matrix should be of shape ({outputs_0.shape[0]}, "
                err += f"{self.state_size}) but is {self.Wout.shape}."
                assert (outputs_0.shape[0], self.state_size) == self.Wout.shape, err

    def _get_next_state(self,
                        single_input: np.ndarray) -> np.ndarray:
        """Given a state vector x(t) and an input vector u(t), compute the state vector x(t+1).

        Parameters:
            single_input {np.ndarray} -- Input vector u(t).

        Raises:
            RuntimeError: feedback is enabled but no feedback vector is available.

        Returns:
            np.ndarray -- Next state s(t+1).
        """

        # check if feedback weights matrix is not None but empty feedback
        if self.Wfb is not None and self.output_values is None:
            raise RuntimeError("Missing a feedback vector.")

        x = self.state[1:self.N+1]

        # add bias
        if self.in_bias:
            u = np.hstack((1, single_input)).astype(self.typefloat)
        else:
            u = single_input

        # linear transformation
        x1 = self.Win @ u.reshape(-1, 1) + self.W @ x

        # add feedback if requested
        if self.Wfb is not None:
            x1 += self.Wfb @ self.fbfunc(self.output_values)

        # previous states memory leak and non-linear transformation
        x1 = (1-self.lr)*x + self.lr*np.tanh(x1)

        # return the next state computed
        if self.use_raw_inp:
            self.state = np.vstack((1.0, x1, u.reshape(-1, 1)))
        else:
            self.state = np.vstack((1.0, x1))

        return self.state.copy()

    def compute_output_from_current_state(self):
        """ Compute output from current state s(t) of the reservoir.

        Returns
        -------
            np.ndarray
                Output at time t.
        """

        assert self.Wout is not None, "Matrix Wout is not initialized/trained yet"

        self.output_values = (self.Wout @ self.state).astype(self.typefloat)
        return self.output_values.copy().ravel()

    def compute_output(self,
                       single_input: np.ndarray,
                       wash_nr_time_step: int = 0):
        """ Compute output from input to the reservoir.

        Parameters
        ---------
            single_input: np.ndarray
                Input vector u(t)
            wash_nr_time_step : int, optional
                Time for reservoir to run without collecting output
                or fitting ``Wout``. (default, 0)

        Returns
        -------
            np.ndarray, np.ndarray
                New state after input u(t) is passed and
                output after input u(t) is passed
        """

        state = self._get_next_state(single_input)
        output = self.compute_output_from_current_state()

        return output, state

    def reset_reservoir(self):
        """Reset reservoir by setting internal values to zero.

        """
        self.state = np.zeros((self.state_size, 1), dtype=self.typefloat)

    def reset_correlation_matrix(self):
        """Reset inverse correlation state matrix to the initial value :

        .. math::

            Corr^{inv} = Id_{N} * \\alpha

        where :math:`\\alpha` is the ``alpha_coef`` and :math:`N` is the
        number of units in the reservoir.
        """
        self.state_corr_inv = np.asmatrix(np.eye(self.state_size)) / self.alpha_coef

    def train_from_current_state(self,
                                 targeted_output: np.ndarray,
                                 indexes: Union[int, list] = None):
        """ Train Wout from current internal state.

        Parameters
        ---------
            targeted_output : np.ndarray
                Expected output of the ESN.
            indexes : int or list, optional
                If indexes is not None, only the provided output
                indexes are learned.
        """

        error = self.output_values - targeted_output.reshape(-1, 1)

        self.state_corr_inv = _new_correlation_matrix_inverse(self.state,
                                                              self.state_corr_inv)

        if indexes is None:
            self.Wout -= error @ (self.state_corr_inv @ self.state).T
        else:
            self.Wout[indexes] -= error[indexes] * (self.state_corr_inv @ self.state).T

    def train(self,
              inputs: Sequence[np.ndarray],
              teachers: Sequence[np.ndarray],
              wash_nr_time_step: int = 0,
              verbose: bool = False) -> Sequence[np.ndarray]:
        """Train the ESN model on a sequence of inputs.

        Parameters
        ---------
            inputs: list of numpy.ndarray
                List of inputs.
                Note that it should always be a list of sequences, i.e. if
                only one sequence (array with rows representing time axis)
                of inputs is used, it should be alone in a list.
            teachers: list of numpy.ndarray
                List of ground truths.
                Note that is should always be a list of
                sequences of the same length than the `inputs`, i.e. if
                only one sequence of inputs is used, it should be alone in a
                list.
            wash_nr_time_step: int, optional
                Number of states to considered as transient when training. Transient
                states will be discarded when computing readout matrix. By default,
                no states are removes.
            verbose : bool

        Returns
        -------
            list of np.ndarray
                All states computed, for all inputs.
        """
        inputs_concat = [inp[t, :] for inp in inputs for t in range(inp.shape[0])]
        teachers_concat = [tea[t, :] for tea in teachers for t in range(tea.shape[0])]

        ## Autochecks of inputs and outputs
        self._autocheck_io(inputs=inputs_concat, outputs=teachers_concat)

        if verbose:
            steps = np.sum([i.shape[0] for i in inputs])
            print(f"Training on {len(inputs)} inputs ({steps} steps) "
                  f"-- wash: {wash_nr_time_step} steps")

        # List of all internal states when training
        all_states = []
        start = 1 if self.in_bias else 0
        end = self.N + start

        for i in range(len(inputs)):

            t = 0
            all_states_inp_i = []

            # First 'warm up' the network
            while t < wash_nr_time_step:
                self.compute_output(inputs_concat[i+t])
                t += 1

            # Train Wout on each input
            while t < inputs[i].shape[0]:
                _, state = self.compute_output(inputs_concat[i+t])
                self.train_from_current_state(teachers_concat[i+t])
                all_states_inp_i.append(state[start:end])
                t += 1

            # Pack in all_states
            all_states.append(np.hstack(all_states_inp_i))

        # return all internal states
        return [st.T for st in all_states]

    def run(self,
            inputs: Sequence[np.ndarray],
            verbose: bool = False) -> Tuple[Sequence[np.ndarray], Sequence[np.ndarray]]:
        """Run the model on a sequence of inputs, and returned the states and
           readouts vectors.

        Parameters
        ----------
            inputs: list of numpy.ndarray
                List of inputs.
                Note that it should always be a list of sequences, i.e. if
                only one sequence (array with rows representing time axis)
                of inputs is used, it should be alone in a list.
            verbose : bool

        Returns
        -------
            list of numpy.ndarray, list of numpy.ndarray
                All outputs computed from readout and all corresponding internal states,
                for all inputs.
        """

        inputs_concat = [inp[t,:] for inp in inputs for t in range(inp.shape[0])]

        steps = np.sum([i.shape[0] for i in inputs])
        if verbose:
            print(f"Running on {len(inputs)} inputs ({steps} steps)")

        # autochecks of inputs
        self._autocheck_io(inputs=inputs_concat)

        all_outputs = []
        all_states = []
        for i in range(len(inputs)):
            internal_pred = []; output_pred = []
            for t in range(inputs[i].shape[0]):
                output, state = self.compute_output(inputs_concat[i+t])
                internal_pred.append(state)
                output_pred.append(output)
            all_states.append(np.asarray(internal_pred))
            all_outputs.append(np.asarray(output_pred))

        # return all_outputs, all_int_states
        return all_outputs, all_states

    def save(self, directory: str):
        """Save the ESN to disk.

        Parameters
        ----------
            directory: str or Path
                Directory where to save the model..
        """
        _save(self, directory)


def _new_correlation_matrix_inverse(new_data, old_corr_mat_inv):
    """
        If old_corr_mat_inv is an approximation for the correlation
        matrix inverse of a dataset (p1, ..., pn), then the function
        returns an approximatrion for the correlation matrix inverse
        of dataset (p1, ..., pn, new_data)

        TODO : add forgetting parameter lbda
    """

    P = old_corr_mat_inv
    x = new_data

    # TODO : numerical instabilities if xTP is not computed first
    # (order of multiplications)
    xTP = x.T @ P
    P = P - (P @ x @ xTP)/(1. + np.dot(xTP, x))

    return P
